from mpl_toolkits.basemap import Basemap
from matplotlib.colors import BoundaryNorm
import matplotlib.pyplot as plt
import numpy as np

from metpy.cbook import get_test_data
from gridding_functions import (remove_nan_observations,
                                remove_repeat_coordinates, estimate_grid_dx, 
                                calc_avg_spacing, define_grid, multi_pass_barnes, interpolate)

from netCDF4 import Dataset


def basic_map():
    """Make our basic default map for plotting"""
    plt.figure(figsize=(17, 11))
    m = Basemap(llcrnrlon=-117, llcrnrlat=21.7, urcrnrlat=50.8, urcrnrlon=-56 ,\
                resolution='l',area_thresh=5000.,projection='lcc',\
                lat_1=80.,lat_2=89,lat_0=35,lon_0=-97.5)

    m.drawcoastlines(color='#A5A5A5')
    m.drawcountries(color='#A5A5A5')
    m.drawstates(color='#A5A5A5')
    m.drawparallels(np.arange(-80.,81.,10.), linestyle='--', linewidth=0.5, color='#B5B5B5', dashes=[30,30])
    m.drawmeridians(np.arange(-180.,181.,10.), linestyle='--', linewidth=0.5, color='#B5B5B5', dashes=[30,30])
    return m

def remove_data_outside_domain(lat, lon, z, m):
    idx = np.where(( lon <= m.lonmax) & ( lon >= m.lonmin ) & 
                   ( lat <= m.latmax) & ( lat >= m.latmin ))[0]
    return lat[idx], lon[idx], z[idx]
    
def load_stations(variable_names, level=500):
    with open('station_data.txt') as f:
        all_data = np.loadtxt(f, skiprows=1, delimiter=',',
                              usecols=(1, 2, 3, 4, 5, 6, 7, 17, 18, 19),
                              dtype=np.dtype([('stid', '3S'), ('lat', 'f'), ('lon', 'f'),
                                              ('slp', 'f'), ('air_temperature', 'f'),
                                              ('cloud_fraction', 'f'), ('dewpoint', 'f'),
                                              ('weather', '16S'),
                                              ('wind_dir', 'f'), ('wind_speed', 'f')]))

    all_stids = [s.decode('ascii') for s in all_data['stid']]

    data = np.concatenate([all_data[all_stids.index(site)].reshape(1, ) for site in all_stids])

    value = data[variable_names]
    lon = data['lon']
    lat = data['lat']

    return lon, lat, value

def loadData(variable_names, level=500):
    with open('2017051900.csv') as f:
        all_data = np.genfromtxt(f, skip_header=1, delimiter=',',
                                 usecols=(0, 2, 3, 5, 6, 7, 8, 9, 10),
                                  dtype=np.dtype([('stid', '3S'), ('lat', 'f'), ('lon', 'f'),
                                                  ('pressure', 'f'), ('air_temperature', 'f'),
                                                  ('dewpoint', 'f'), ('height', 'f'),
                                                  ('wind_dir', 'f'), ('wind_speed', 'f')]),
                                invalid_raise=False,
                                missing_values='--')
    all_stids = [s.decode('ascii') for s in all_data['stid']]
    all_pressures = all_data['pressure']
    idx_match_pressure = np.where(all_pressures == level)

    #data = np.concatenate([all_data[all_stids.index(site)].reshape(1, ) for site in all_stids])
    value = all_data[variable_names][idx_match_pressure]
    lon = all_data['lon'][idx_match_pressure]
    lat = all_data['lat'][idx_match_pressure]
    
    return lon, lat, value

cmap = plt.get_cmap('magma')

# Load in the observations
lon, lat, height = loadData('height')
m = basic_map()
lat, lon, height = remove_data_outside_domain(lat, lon, height, m)
plt.close()

x, y = m(lon, lat)
x, y, height = remove_nan_observations(x, y, height)
x, y, height = remove_repeat_coordinates(x, y, height)

avg_data_spacing = calc_avg_spacing(x,y)

dx = 40000 # 12 km grid...similar to NAM
grid = define_grid(x, y, dx, 0, 0, 0, 0, 0)
print "Grid size:", grid[0].shape

#Barnes Interpolation With First Guess
nam_first_guess_data = Dataset('nam_218_20170518_1800_006.grb2.nc')
nam_lat, nam_lon = nam_first_guess_data.variables['lat'][:], nam_first_guess_data.variables['lon'][:]

# Load in NAM data and NAM grid
nam_x, nam_y = m(nam_lon, nam_lat)
isobaric = nam_first_guess_data.variables['isobaric1'][:]/100.
plevel_idx = np.where(isobaric == 500)[0]
stride = 2
nam_height = (nam_first_guess_data.variables['Geopotential_height_isobaric'][0,plevel_idx,:,:].squeeze())[::stride,::stride]

nam_grid = (nam_x[::stride,::stride], nam_y[::stride,::stride])

print "Interpolating NAM Grid to new analysis grid."
# interpolate NAM grid to analysis grid
new_nam_hght = interpolate(nam_x[::stride,::stride].flatten(), nam_y[::stride,::stride].flatten(), nam_height.flatten(), None, grid=grid, interp_type='linear')[2]
print "Done with interpolation."

m = basic_map()
# Perform the multi-pass Barnes analysis
img = multi_pass_barnes(np.asarray(x), np.asarray(y), np.asarray(height), 
                        grid, interp_type='barnes', ave_spacing=avg_data_spacing, gamma=0.2,
                        search_radius=10000000, passes=2, guess=new_nam_hght)

img = np.ma.masked_where(np.isnan(img), img)
mmb = m.contourf(grid[0], grid[1], img, np.arange(5360,6030,30), cmap=cmap)
sc = plt.scatter(x, y, c=height, vmin=5360, vmax=6000, cmap=cmap, edgecolor='k')
plt.colorbar(mmb, shrink=0.8, pad=0)
plt.colorbar(sc, shrink=0.8, pad=0)
plt.show()